from overrides import override

from vortezwohl.nlp.base_string_similarity import BaseStringSimilarity


class LevenshteinDistance(BaseStringSimilarity):
    def __init__(self, ignore_case: bool = False):
        super().__init__(ignore_case=ignore_case)

    @override
    def __call__(self, s_1: str, s_2: str) -> int:
        if self._ignore_case:
            s_1, s_2 = s_1.lower(), s_2.lower()
        m, n = len(s_1), len(s_2)
        dp = [[0] * (n + 1) for _ in range(m + 1)]
        for i in range(m + 1):
            dp[i][0] = i
        for j in range(n + 1):
            dp[0][j] = j
        for i in range(1, m + 1):
            for j in range(1, n + 1):
                if s_1[i - 1] == s_2[j - 1]:
                    dp[i][j] = dp[i - 1][j - 1]
                else:
                    dp[i][j] = min(
                        dp[i - 1][j - 1] + 1,
                        dp[i - 1][j] + 1,
                        dp[i][j - 1] + 1
                    )
        return dp[m][n]

    @override
    def rank(self, s: str, S: list[str]) -> list[str]:
        lcs_list = sorted([(self.__call__(s, x), x) for x in S], key=lambda x: x[0], reverse=False)
        return [x[1] for x in lcs_list]



__AUTHOR__ = '吴子豪 / Vortez Wohl'
__EMAIL__ = 'vortez.wohl@gmail.com'

import logging

NEW_LINE = '\n'
BLANK = ' '
UTF_8 = 'utf-8'

logger = logging.getLogger('vortezwohl')
logger.setLevel(logging.INFO)
formatter = logging.Formatter('[%(levelname)s] %(asctime)s %(name)s : %(message)s')
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logger = logging.getLogger('vortezwohl.io')
logger.setLevel(logging.INFO)
logger = logging.getLogger('vortezwohl.retry')
logger.setLevel(logging.INFO)



import logging
import random
import time

from typing_extensions import Callable, Any

from vortezwohl import NEW_LINE, BLANK

logger = logging.getLogger('vortezwohl.retry')


def sleep(retries: int, base: float = 2., max_delay: float = 600.):
    retries = max(retries, 1)
    delay = min(base ** retries, max_delay)
    time.sleep(delay + random.uniform(.1, delay))
    return


class MaxRetriesReachedError(RuntimeError):
    def __init__(self, retries: int, message: str):
        super().__init__(f'Max attempts reached: {retries + 1}\n{message}')


class Retry:
    def __init__(self, max_retries: int | None = None, delay: bool = False):
        self._max_retries = max_retries
        self._delay = delay

    def on_return(self, validator: Callable[Any, bool]):
        def decorator(func: Callable):
            def wrapper(*_args, **_kwargs):
                result = func(*_args, **_kwargs)
                if validator(result):
                    return result
                else:
                    logger.debug('Validation failed.\n'
                                 '- Retrying...')
                    if self._max_retries is None:
                        retry_count = 0
                        while True:
                            if self._delay:
                                sleep(retries=retry_count)
                            result = func(*_args, **_kwargs)
                            if validator(result):
                                return result
                            else:
                                logger.debug('Validation failed.\n'
                                             f'- Retry {retry_count + 1}/inf'
                                             f' : {str(result).replace(NEW_LINE, BLANK)}')
                    else:
                        for retry_count in range(self._max_retries):
                            if self._delay:
                                sleep(retries=retry_count)
                            result = func(*_args, **_kwargs)
                            if validator(result):
                                return result
                            else:
                                logger.debug(f'Validation failed.\n'
                                             f'- Retry {retry_count + 1}/{self._max_retries}'
                                             f' : {str(result).replace(NEW_LINE, BLANK)}')
                        raise MaxRetriesReachedError(retries=self._max_retries,
                                                     message='Validation failed.\n'
                                                             f'Returns: {result}')
            return wrapper
        return decorator

    def on_exceptions(self, exceptions: list[type] | type):
        if isinstance(exceptions, type):
            exceptions = [exceptions]

        def decorator(func: Callable):
            def wrapper(*_args, **_kwargs):
                def validator() -> tuple[tuple, bool, Any]:
                    result = None
                    need_retry = False
                    error_type = None
                    error_super_type = None
                    error = None
                    try:
                        result = func(*_args, **_kwargs)
                    except Exception as e:
                        for exception in exceptions:
                            if isinstance(e, exception):
                                need_retry = True
                                error = e
                                error_type = e.__class__.__name__
                                error_super_type = exception.__name__
                                break
                    return (error, error_type, error_super_type), need_retry, result
                (_error, _error_type, _error_super_type), _need_retry, _result = validator()
                if not _need_retry:
                    return _result
                else:
                    logger.debug(f'Validation failed, {_error_type}({_error_super_type}) occurred: {str(_error)}.\n'
                                 '- Retrying...')
                    if self._max_retries is None:
                        retry_count = 0
                        while True:
                            if self._delay:
                                sleep(retries=retry_count)
                            (_error, _error_type, _error_super_type), _need_retry, _result = validator()
                            if not _need_retry:
                                return _result
                            else:
                                logger.debug(f'Validation failed, {_error_type}({_error_super_type}) occurred: {str(_error)}.\n'
                                             f'- Retry {retry_count + 1}/inf '
                                             f': {str(_result).replace(NEW_LINE, BLANK)}')
                    else:
                        for retry_count in range(self._max_retries):
                            if self._delay:
                                sleep(retries=retry_count)
                            (_error, _error_type, _error_super_type), _need_retry, _result = validator()
                            if not _need_retry:
                                return _result
                            else:
                                logger.debug(f'Validation failed, {_error_type}({_error_super_type}) occurred: {str(_error)}.\n' 
                                             f'- Retry {retry_count + 1}/{self._max_retries} '
                                             f': {str(_result).replace(NEW_LINE, BLANK)}')
                        raise MaxRetriesReachedError(retries=self._max_retries,
                                                     message=f'{_error_type}({_error_super_type}) '
                                                             f'occurred: {str(_error)}\n'
                                                             f'Returns: {_result}')
            return wrapper
        return decorator



import threading
from typing_extensions import Callable, Any


def timeout(_timeout: float, _default: Any = None):
    def decorator(func: Callable):
        def wrapper(*_args, **_kwargs):
            result = []
            err = []

            def __wrapper():
                try:
                    result.append(func(*_args, **_kwargs))
                except Exception as e:
                    err.append(e)
                return
            thread = threading.Thread(target=__wrapper)
            thread.start()
            thread.join(timeout=_timeout)
            if thread.is_alive():
                if _default is None:
                    raise TimeoutError(f'{func} timed out after {_timeout} seconds.')
                else:
                    return _default
            if len(err) > 0:
                raise err[0]
            return result[0] if len(result) > 0 else None
        return wrapper
    return decorator



from typing_extensions import Iterable, Any


def batched(_iterable: Iterable, batch_size: int = 1, fill_value: Any = None) -> list:
    step = 0
    tmp_iter = []
    for item in _iterable:
        step += 1
        tmp_iter.append(item)
        if step % batch_size == 0:
            yield tmp_iter.copy()
            tmp_iter.clear()
    if len(tmp_iter) > 0:
        while len(tmp_iter) < batch_size:
            tmp_iter.append(fill_value)
        yield tmp_iter


def sliding_window(_iterable: Iterable, window_size: int, stride: int, fill_value: Any = None) -> Iterable:
    _list_iterable = list(_iterable)
    outer_step = 0
    for _ in _list_iterable:
        if outer_step % stride == 0:
            tmp_iter = []
            inner_step = 0
            for item in _list_iterable[outer_step:]:
                inner_step += 1
                tmp_iter.append(item)
                if inner_step % window_size == 0 and inner_step > 0:
                    yield tmp_iter.copy()
                    tmp_iter.clear()
                    break
            if len(tmp_iter) > 0:
                while len(tmp_iter) < window_size:
                    tmp_iter.append(fill_value)
                yield tmp_iter.copy()
        outer_step += 1



import traceback
from concurrent.futures import ThreadPoolExecutor, Future, as_completed
from threading import RLock
from types import NoneType
from typing_extensions import Iterable, Callable, Dict, Tuple

import psutil

from vortezwohl.concurrent.future_result import FutureResult


class ThreadPool:
    def __init__(self, max_workers: int | None = None):
        if max_workers is None:
            max_workers = psutil.cpu_count(logical=True) + 1
        self._max_workers = max_workers
        self._thread_pool_executor = ThreadPoolExecutor(max_workers=self._max_workers)
        self._futures = list()
        self._futures_lock = RLock()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.shutdown(cancel_futures=False)

    def submit(self, job: Callable, *args, **kwargs) -> Future:
        _future = None
        if len(args) > 0:
            _future = self._thread_pool_executor.submit(lambda param: (job, param, job(*param)), args)
        elif len(kwargs) > 0:
            _future = self._thread_pool_executor.submit(lambda param: (job, param, job(**param)), kwargs)
        else:
            _future = self._thread_pool_executor.submit(lambda: (job, None, job()))
        with self._futures_lock:
            self._futures.append(_future)
        return _future

    def gather(self, jobs: Iterable[Callable], arguments: list[Dict] | list[Tuple] | list[NoneType] | None = None) -> Iterable[FutureResult]:
        futures = []
        for i, job in enumerate(jobs):
            if arguments is None or arguments[i] is None:
                futures.append(self._thread_pool_executor.submit(lambda: (job, None, job())))
                continue
            if isinstance(arguments[i], dict):
                futures.append(self._thread_pool_executor.submit(lambda param: (job, param, job(**param)), arguments[i]))
            elif isinstance(arguments[i], Iterable):
                futures.append(self._thread_pool_executor.submit(lambda param: (job, param, job(*param)), arguments[i]))
            else:
                futures.append(self._thread_pool_executor.submit(lambda param: (job, param, job(param)), arguments[i]))
        for future in as_completed(futures):
            try:
                res = future.result()
                yield FutureResult(_callable=res[0], arguments=res[1], returns=res[2],
                                   error=None, traceback=None)
            except Exception as __e:
                yield FutureResult(_callable=None, arguments=None, returns=None,
                                   error=__e, traceback=traceback.format_exc())

    def map(self, job: Callable, arguments: list[Dict] | list[Tuple] | list[NoneType] | None = None) -> Iterable[FutureResult]:
        return self.gather(jobs=[job] * len(arguments), arguments=arguments)

    def shutdown(self, cancel_futures: bool = True):
        self._thread_pool_executor.shutdown(wait=True, cancel_futures=cancel_futures)
        return self

    def wait_all(self) -> list[FutureResult]:
        _res = []
        for _ in self.next_result:
            _res.append(_)
        return _res

    @property
    def next_result(self) -> Iterable[FutureResult]:
        for f in as_completed(self._futures):
            try:
                res = f.result()
                yield FutureResult(_callable=res[0], arguments=res[1], returns=res[2],
                                   error=None, traceback=None)
            except Exception as __e:
                yield FutureResult(_callable=None, arguments=None, returns=None,
                                   error=__e, traceback=traceback.format_exc())
            finally:
                with self._futures_lock:
                    self._futures.remove(f)

    @property
    def thread_pool_executor(self) -> ThreadPoolExecutor:
        return self._thread_pool_executor

    @property
    def max_workers(self) -> int:
        return self._max_workers



import hashlib


def md5(text: str) -> str:
    _hash = hashlib.md5()
    text_bytes = text.encode('utf-8')
    _hash.update(text_bytes)
    return _hash.hexdigest()


def sha1(text: str) -> str:
    _hash = hashlib.sha1()
    text_bytes = text.encode('utf-8')
    _hash.update(text_bytes)
    return _hash.hexdigest()


def sha256(text: str) -> str:
    _hash = hashlib.sha256()
    text_bytes = text.encode('utf-8')
    _hash.update(text_bytes)
    return _hash.hexdigest()


def sha512(text: str) -> str:
    _hash = hashlib.sha512()
    text_bytes = text.encode('utf-8')
    _hash.update(text_bytes)
    return _hash.hexdigest()



import threading

from typing_extensions import Any
from overrides import override

from vortezwohl.cache.base_cache import BaseCache


class LRUCache(BaseCache):
    def __init__(self, capacity: int):
        super().__init__()
        self._capacity = capacity
        self._recently_used = list()
        self._recently_used_lock = threading.RLock()

    def expand_to(self, capacity: int) -> None:
        self._capacity = capacity

    def __recently_used_enqueue(self, key: Any) -> Any:
        with self._recently_used_lock:
            if key in self._recently_used:
                self._recently_used.remove(key)
            self._recently_used.append(key)
            while len(self._recently_used) > self._capacity:
                self.delete(self.__recently_used_dequeue())
        return key

    def __recently_used_dequeue(self) -> None:
        key = None
        with self._recently_used_lock:
            if len(self._recently_used) > 0:
                key = self._recently_used.pop(0)
        return key

    @override
    def read(self, key: Any) -> Any | None:
        self.__recently_used_enqueue(key=key)
        return super().read(key=key)

    @override
    def write(self, key: Any, value: Any) -> Any | None:
        self.__recently_used_enqueue(key=key)
        return super().write(key=key, value=value)



import logging
import requests

from vortezwohl import NEW_LINE, BLANK, UTF_8
from vortezwohl.func.retry import sleep

logger = logging.getLogger('vortezwohl.io')


class HttpClient:
    def __init__(self, max_retries: int = 3, timeout: float = 8., _delay_base: float = 2.):
        self._max_retries = max_retries
        self._timeout = timeout
        self._delay_base = _delay_base

    @staticmethod
    def sleep(retries: int, base: float = 2., max_delay: float = 600.):
        return sleep(retries=retries, base=base, max_delay=max_delay)

    def get(self, url: str, data: dict | None = None, headers: dict | None = None, **kwargs):
        r = None
        retry_count = 0
        for _ in range(self._max_retries + 1):
            if r is not None:
                retry_count += 1
                _content = r.content.replace(NEW_LINE.encode(UTF_8), BLANK.encode(UTF_8)).strip().decode(UTF_8)
                logger.debug(f'({r.status_code}) {url} Retry {retry_count}/{self._max_retries}'
                             + (f' : {_content}' if len(_content) > 0 else ''))
                if retry_count >= self._max_retries:
                    return r
                self.sleep(_, base=self._delay_base, **kwargs)
            r = requests.get(url=url, params=data, headers=headers, timeout=self._timeout)
            if r.status_code == 200:
                return r
        return r

    def post(self, url: str, data: dict, headers: dict | None = None, **kwargs):
        r = None
        retry_count = 0
        for _ in range(self._max_retries + 1):
            if r is not None:
                retry_count += 1
                _content = r.content.replace(NEW_LINE.encode(UTF_8), BLANK.encode(UTF_8)).strip().decode(UTF_8)
                logger.debug(f'({r.status_code}) {url} Retry {retry_count}/{self._max_retries}'
                             + (f' : {_content}' if len(_content) > 0 else ''))
                if retry_count >= self._max_retries:
                    return r
                self.sleep(_, base=self._delay_base, **kwargs)
            r = requests.post(url=url, json=data, headers=headers, timeout=self._timeout)
            if r.status_code == 200:
                return r
        return r



import threading

from typing_extensions import Any, OrderedDict


class BaseCache:
    def __init__(self):
        self._cache = OrderedDict()
        self._cache_lock = threading.RLock()

    def read(self, key: Any) -> Any | None:
        value = None
        with self._cache_lock:
            value = self._cache.get(key, None)
        return value

    def write(self, key: Any, value: Any) -> Any | None:
        with self._cache_lock:
            self._cache[key] = value
        return value

    def flush(self) -> None:
        with self._cache_lock:
            self._cache.clear()
        return

    def delete(self, key: Any) -> Any | None:
        with self._cache_lock:
            if key in self._cache.keys():
                value = self._cache[key]
                del self._cache[key]
                return value
        return

    def __contains__(self, key: Any) -> bool:
        contains = False
        with self._cache_lock:
            contains = (key in self._cache.keys())
        return contains

    def __getitem__(self, key: Any) -> Any:
        return self.read(key=key)

    def __setitem__(self, key: Any, value: Any) -> Any | None:
        return self.write(key=key, value=value)

    def __delitem__(self, key: Any) -> Any | None:
        return self.delete(key=key)

    def __str__(self) -> str:
        stringify = ''
        with self._cache_lock:
            _max_len = 8
            for key, value in self._cache.items():
                _max_len = max(_max_len, len(str(key)))
                _max_len = max(_max_len, len(str(value)))
            _padding = 2
            _max_len += _padding
            line_spliter = f'\n{"+" + "=" * (_max_len * 2 + 1) + "+"}\n'
            header = f'{self.__class__.__name__}(capacity={len(self._cache.keys())})'
            stringify = (f'{header:^{_max_len * 2 + 2}}{line_spliter}'
                         + f'|{"Key":^{_max_len}}|{"Value":^{_max_len}}|{line_spliter}'
                         + line_spliter.join([f'|{str(k):^{_max_len}}|{str(v):^{_max_len}}|' for k, v in self._cache.items()])
                         + (line_spliter if len(self) else ''))
        return stringify

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        length = 0
        with self._cache_lock:
            length = len(self._cache.keys())
        return length

    def __eq__(self, other) -> bool:
        stringify = ''
        with self._cache_lock:
            stringify = str(self)
        return stringify == str(other)

    def __call__(self, key: Any) -> Any:
        return self.read(key=key)



import re
import os
import sys
import subprocess

from vortezwohl.io.file import get_files, size_of
from vortezwohl.nlp.levenshtein_distance import LevenshteinDistance

lev_distance = LevenshteinDistance(ignore_case=True)


class System:
    def __init__(self):
        self._python = sys.executable
        self._venv_executable = os.path.dirname(self._python)
        self._workdir = os.getcwd()
        self.shell(f'{self._python} -m ensurepip --upgrade', force_check=True)
        __pip_name = lev_distance.rank('pip', [os.path.basename(_) for _ in get_files(self._venv_executable)])[0]
        self._pip = os.path.join(self._venv_executable, __pip_name)

    def shell(self, cmd: str, workdir: str | None = None, timeout: int | None = None,
              force_check: bool = False) -> tuple[str, int, str]:
        if workdir is None:
            workdir = self._workdir
        res = subprocess.run(cmd, cwd=workdir, shell=True, capture_output=True, text=True,
                             timeout=timeout, check=force_check)
        return res.stdout, res.returncode, res.stderr

    def script(self, script: str, argument: str):
        return self.shell(f'{os.path.join(self._venv_executable, script)} {argument}')

    def interpret(self, script: str, is_module: bool = False):
        if is_module:
            return self.shell(f'{self._python} -m {script}', force_check=False)
        return self.shell(f'{self._python} {script}', force_check=False)

    def list_packages(self):
        tmp_res = self.shell(f'{self._pip} list', force_check=True)[0].splitlines(keepends=False)[2:]
        res = []
        blank_pattern = re.compile(r'(\s+)')
        for line in tmp_res:
            if len(line.strip()) < 1:
                break
            res.append(line)
        for i, line in enumerate(res):
            _tmp_line = re.split(pattern=blank_pattern, string=line)
            res[i] = f'{_tmp_line[0]}=={_tmp_line[2]}'
        return res

    def show_package(self, package: str) -> list[str]:
        try:
            return self.shell(f'{self._pip} show {package}', force_check=True)[0].splitlines(keepends=False)
        except subprocess.CalledProcessError:
            return []

    def package_size(self, package: str) -> int:
        location = [line for line in self.show_package(package) if line.lower().__contains__('location:')]
        if len(location) > 0:
            location = location[0]
            location = location[location.find(':') + 1:].strip()
            return size_of(location)
        else:
            return 0

    def package_version(self, package: str) -> str:
        version = [line for line in self.show_package(package) if line.lower().__contains__('version:')]
        if len(version) > 0:
            version = version[0]
            return version[version.find(':') + 1:].strip()
        else:
            return 'Unknown'

    def install_package(self, package: str, version: str | None = None) -> str | bool:
        try:
            if version is not None:
                self.shell(f'{self._pip} install {package}~={version}', force_check=True)
            else:
                self.shell(f'{self._pip} install --upgrade {package}', force_check=True)
            return f'{package}=={self.package_version(package)}'
        except subprocess.CalledProcessError:
            return False

    def uninstall_package(self, package: str) -> str | bool:
        try:
            version = self.package_version(package)
            subprocess.run(f'{self._pip} uninstall {package}', cwd=self._workdir,
                           shell=True, capture_output=True, text=True,
                           timeout=None, check=True, input='Y')
            return f'{package}=={version}'
        except subprocess.CalledProcessError:
            return False



import threading
from typing_extensions import Callable, Union


def lock_on(lock: Union[threading.Lock, threading.RLock, threading.Semaphore], **kwargs):
    def decorator(func: Callable):
        def wrapper(*_args, **_kwargs):
            if 'blocking' not in kwargs or kwargs.get('timeout', None) is not None:
                kwargs['blocking'] = True
            try:
                lock.acquire(**kwargs)
                return func(*_args, **_kwargs)
            finally:
                if isinstance(lock, threading.Semaphore):
                    lock.release(n=1)
                else:
                    lock.release()
        return wrapper
    return decorator



import os
import threading
import random

SEED_LOCK = threading.RLock()
SEED_INCREMENT = None
SEED_MULTIPLIER = None
MOD = int(1e64)


def next_seed() -> int:
    global SEED_INCREMENT, SEED_MULTIPLIER
    SEED_INCREMENT = int(os.getenv('INITIAL_SEED', 0)) if SEED_INCREMENT is None else SEED_INCREMENT
    SEED_MULTIPLIER = int(os.getenv('SEED_MULTIPLIER', 1)) if SEED_MULTIPLIER is None else SEED_MULTIPLIER
    SEED_LOCK.acquire(blocking=True)
    SEED_INCREMENT = ((SEED_INCREMENT + 1) * SEED_MULTIPLIER * max((SEED_INCREMENT // MOD), 1)) % MOD
    SEED_MULTIPLIER += 1
    SEED_LOCK.release()
    random.seed(SEED_INCREMENT)
    return SEED_INCREMENT



from typing_extensions import Callable, Dict, Tuple, Any


class FutureResult:
    def __init__(self, _callable: Callable | None, arguments: Dict | Tuple | None,
                 returns: Any, error: Exception | None, traceback: str | None):
        self._callable = _callable
        self._arguments = arguments
        self._returns = returns
        self._error = error
        self._traceback = traceback

    @property
    def callable(self):
        return self._callable

    @property
    def arguments(self):
        return self._arguments

    @property
    def returns(self):
        return self._returns

    @property
    def error(self):
        return self._error

    @property
    def traceback(self):
        return self._traceback



from abc import abstractmethod

from typing_extensions import Callable


class BaseStringSimilarity(Callable):
    def __init__(self, ignore_case: bool, *args, **kwargs):
        self._ignore_case = ignore_case

    @abstractmethod
    def __call__(self, *args, **kwargs): ...

    @abstractmethod
    def rank(self, s: str, S: list[str]) -> list[str]: ...



import os


def read_file(path: str, encoding: str = 'utf-8') -> str:
    try:
        with open(path, mode='r', encoding=encoding) as f:
            return f.read()
    except UnicodeDecodeError:
        try:
            with open(path, mode='r', encoding='gbk') as f:
                return f.read()
        except UnicodeDecodeError:
            with open(path, mode='r', encoding=encoding, errors='replace') as f:
                return f.read()


def write_file(content: str | bytes, path: str, encoding: str = 'utf-8') -> str:
    os.makedirs(os.path.dirname(path), exist_ok=True)
    if isinstance(content, bytes):
        with open(path, mode='wb') as f:
            f.write(content)
        return path
    with open(path, mode='w', encoding=encoding) as f:
        f.write(content)
    return os.path.abspath(path)


def get_files(path: str) -> list:
    path = os.path.abspath(path)
    if os.path.exists(path):
        entries = os.listdir(path)
        return [os.path.join(path, entry) for entry in entries if os.path.isfile(os.path.join(path, entry))]
    else:
        return []


def size_of(path: str) -> int:  # bytes
    def size_of_dir(dir_path: str) -> int:
        total_size = 0
        if not os.path.exists(dir_path) or not os.path.isdir(dir_path):
            return 0
        for item in os.listdir(dir_path):
            item_path = os.path.join(dir_path, item)
            if os.path.isfile(item_path):
                total_size += os.path.getsize(item_path)
            elif os.path.isdir(item_path):
                total_size += size_of_dir(item_path)
        return total_size

    if not os.path.exists(path):
        return 0
    if os.path.isfile(path):
        return os.path.getsize(path)
    elif os.path.isdir(path):
        return size_of_dir(path)
    else:
        raise TypeError(f'The path "{path}" is not a file or directory.')
